from os import path
from pathlib import Path
from typing import Optional, Union

from notifypy import Notify

from reviews import config
from reviews.notifications.models import PullRequestNotification, Sound


class NotificationClient:
    """Notifiction Client provides a warpper and convience methods
    for sending desktop notifications.
    """

    def __init__(self) -> None:
        self.notification = Notify()
        self.application_name = "Code Review Manager"

    @staticmethod
    def _asset_exists(asset: str) -> Union[str, None]:
        """"Checks if an asset exists and returns the path (if it does)"""
        if asset_path := Path(path.join(config.DATA_PATH, asset)):
            if asset_path.exists():
                return str(asset_path)
        return None

    def _send_notification(
        self,
        title: str,
        message: str,
        icon: Optional[str] = None,
        audio: Optional[str] = None,
    ):
        """a generic wrapper around send notification."""
        self.notification.application_name = self.application_name
        self.notification.title = title
        self.notification.message = message
        if icon:
            if file_path := self._asset_exists(asset=icon):
                self.notification.icon = file_path
        if audio:
            if file_path := self._asset_exists(asset=audio):
                self.notification.audio = file_path

        self.notification.send(block=False)

    def send_pull_request_review(self, model: PullRequestNotification) -> bool:
        """sends a notification for a pull request review"""
        return self._send_notification(
            title="Pull request review",
            message=f"New review for {model.name} in {model.org}/{model.repository}",
            audio=Sound.SUCCESS.value,
        )

    def send_pull_request_approved(self, model: PullRequestNotification):
        """sends a notification for a pull request approved"""
        self._send_notification(
            title="Pull request approved",
            message=(
                f"{model.number} has been approved in {model.org}/{model.repository}"
            ),
            audio=Sound.SUCCESS.value,
        )

    def send_pull_request_merged(self, model: PullRequestNotification):
        """sends a notification for a pull request merged"""
        self._send_notification(
            title="Pull request merged",
            message=f"{model.name} has been merged in {model.org}/{model.repository}",
            audio=Sound.FAILURE.value,
        )



from dataclasses import dataclass, field
from datetime import datetime
from typing import List


@dataclass
class Label:
    """Labels on a Pull Request."""

    name: str


@dataclass
class PullRequest:
    """Models a Pull Request"""

    number: int
    title: str
    created_at: datetime
    updated_at: datetime
    approved: bool
    labels: List[Label] = field(default_factory=list)



from reviews.datasource.client import SQLClient  # NOQA: F401
from reviews.datasource.managers import PullRequestManager  # NOQA: F401
from reviews.datasource.models import Label, PullRequest  # NOQA: F401



from typing import Any, List

from reviews.datasource.client import SQLClient
from reviews.datasource.models import PullRequest


class PullRequestManager:
    """Manager for PullRequests Table"""

    client: SQLClient
    table_name: str

    def __init__(self, client: SQLClient) -> None:
        self.client = client
        self.table_name = "pull_requests"

    def create_table(self):
        """creates the table."""
        sql = f"""
            CREATE TABLE IF NOT EXISTS {self.table_name}
             (
                id INTEGER PRIMARY KEY autoincrement,
                number INTEGER,
                title text NOT NULL,
                created_at text NOT NULL,
                updated_at text NOT NULL,
                approved INTEGER DEFAULT 0
             );
            """
        return self.client.query(sql=sql)

    def drop_table(self):
        """drops the Table."""
        sql = f"""
            DROP TABLE IF EXISTS {self.table_name};
            """
        return self.client.query(sql=sql)

    def all(self):
        """fetch all persisted database rows for pull requests."""
        sql = f"""
            SELECT *
            FROM {self.table_name};
        """
        return self.client.query(sql=sql)

    def get_by_id(self, row_id: int):
        """fetch all persisted database rows for pull requests."""
        sql = f"""
            SELECT *
            FROM   {self.table_name}
            WHERE  id = ?;
        """
        return self.client.query(sql=sql, data=(row_id,))

    def insert_all(self, models: List[PullRequest]) -> List[Any]:
        """inserts a list of models into the database"""
        inserted = []
        for model in models:
            row = self.insert(model=model)
            inserted.append(row)

        return inserted

    def insert(self, model):
        """insert pull request into database."""
        sql = f"""
            INSERT INTO {self.table_name}
            (
                    title,
                    created_at,
                    updated_at,
                    approved,
                    number
            )
            VALUES (?,?,?,?,?);
            """

        return self.client.insert(
            sql=sql,
            data=(
                model.title,
                model.created_at,
                model.updated_at,
                model.approved,
                model.number,
            ),
        )

    def update(self, row_id: int, model: PullRequest):
        """upsert pull request into database."""
        sql = f"""
            UPDATE OR REPLACE {self.table_name}
            SET
                title = ?,
                created_at = ?,
                updated_at = ?,
                approved = ?,
                number = ?
            WHERE id = ?;
            """

        return self.client.query(
            sql=sql,
            data=(
                model.title,
                model.created_at,
                model.updated_at,
                model.approved,
                model.number,
                row_id,
            ),
        )

    def delete(self, row_id):
        """delete from the table"""
        sql = f"""
            DELETE FROM {self.table_name}
            WHERE id = ?;
        """
        return self.client.query(
            sql=sql,
            data=(row_id,),
        )

    def exists(self) -> bool:
        """checks if the table exists"""
        sql = f"""
            SELECT count(name) FROM sqlite_master WHERE type='table' AND name='{self.table_name}' LIMIT 1;
        """
        result = self.client.query(
            sql=sql,
        )

        # first element in the array, first result in the tuple
        exists = result[0][0]

        return bool(int(exists))



from rich.console import Console
from rich.panel import Panel

from reviews.layout.components import Header

console = Console()


class RenderLayoutManager:
    """Renders the Layout for the Terminal UI"""

    def __init__(self, layout):
        self.layout = layout

    def render_layout(
        self, progress_table, body, pull_request_component, log_component
    ):
        """Renders the entire layout"""
        self.render_header()
        self.render_body(component=body)
        self.render_log(component=log_component)
        self.render_configuration(component=pull_request_component)
        self.render_footer(progress_table=progress_table)

        return self.layout

    def render_header(self):
        """Renders the Header component"""
        self.layout["header"].update(Header())

    def render_body(self, component):
        """Renders the Body component"""
        self.layout["body"].update(component)

    def render_configuration(self, component):
        """Renders the Main Body component"""
        self.layout["configuration"].update(
            Panel(component, title="Configuration", border_style="blue")
        )

    def render_log(self, component):
        """Rends the log component"""
        self.layout["log"].update(Panel(component, title="Log", border_style="blue"))

    def render_review(self, component):
        """Renders the Review Side Body component"""
        self.layout["review"].update(
            Panel(component, title="Ready to Review", border_style="blue")
        )

    def render_shippable(self, component):
        """Renders the Shippable Side Body component"""
        self.layout["ship"].update(
            Panel(component, title="Ready to Ship", border_style="blue")
        )

    def render_footer(self, progress_table):
        """Renders the Footer component"""
        self.layout["footer"].update(progress_table)



from datetime import datetime
from typing import Dict, List, Tuple

import humanize
from rich import box
from rich.console import RenderGroup
from rich.layout import Layout
from rich.panel import Panel
from rich.progress import BarColumn, Progress, SpinnerColumn, TextColumn
from rich.table import Table
from rich.tree import Tree

from reviews.datasource import PullRequest


def render_pull_request_table(title: str, pull_requests: List[PullRequest]) -> Table:
    """Renders a list of pull requests as a table"""
    table = Table(show_header=True, header_style="bold white")
    table.add_column("#", style="dim", width=5)
    table.add_column(title, width=60)
    table.add_column("Labels", width=40)
    table.add_column("Activity", width=15)
    table.add_column("Status", width=10)

    pull_requests = sorted(pull_requests, key=lambda x: x.updated_at, reverse=True)

    for pr in pull_requests:
        approved = "[green]Approved" if pr.approved else ""
        updated_at = humanize.naturaltime(pr.updated_at)

        colour = ""
        if (datetime.now() - pr.updated_at).days >= 7:
            colour = "[red]"
        elif (datetime.now() - pr.updated_at).days >= 1:
            colour = "[yellow]"
        updated_at = f"{colour}{updated_at}"

        labels = ", ".join([label.name for label in pr.labels])

        table.add_row(
            f"[white]{pr.number} ",
            f"[white]{pr.title}",
            f"{labels}",
            f"{updated_at}",
            f"{approved}",
        )

    return table


def generate_layout() -> Layout:
    """Define the layout for the terminal UI."""
    layout = Layout(name="root")

    layout.split(
        Layout(name="header", size=3),
        Layout(name="main", ratio=1),
        Layout(name="footer", size=7),
    )
    layout["main"].split_row(
        Layout(name="left_side"),
        Layout(name="body", ratio=2, minimum_size=60),
    )
    layout["left_side"].split(Layout(name="configuration"), Layout(name="log"))
    return layout


def generate_tree_layout(configuration: List[Tuple[str, str]]) -> RenderGroup:
    """Generates a tree layout for the settings configuration"""
    organization_tree_mapping: Dict[str, Tree] = {}
    for (org, repo) in configuration:
        tree = organization_tree_mapping.get(f"{org}", Tree(f"[white]{org}"))
        tree.add(f"[link=https://www.github.com/{org}/{repo}]{repo}[/link]")
        organization_tree_mapping[org] = tree

    return RenderGroup(*organization_tree_mapping.values())


def generate_log_table(logs):
    """Generetes a table for logging activity"""
    table = Table("Time", "Message", box=box.SIMPLE)

    for log in logs:
        time, message = log
        table.add_row(time, message)

    return table


def generate_progress_tracker():
    """Tracks the progress of background tasks"""
    progress = Progress(
        "{task.description}",
        SpinnerColumn(),
        BarColumn(),
        TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
    )
    progress.add_task("[white]Pull Requests", total=100)

    total = sum(task.total for task in progress.tasks)
    overall_progress = Progress()
    overall_task = overall_progress.add_task("All", total=int(total))

    progress_table = Table.grid(expand=True)
    progress_table.add_row(
        Panel(
            overall_progress,
            title="Next Refresh",
            border_style="blue",
        ),
        Panel(
            progress, title="[b]Next fetch for:", border_style="blue", padding=(1, 2)
        ),
    )

    return progress, overall_progress, overall_task, progress_table



from setuptools import find_namespace_packages, setup


def _requires_from_file(filename):
    return open(filename).read().splitlines()


with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
    name="reviews",
    packages=find_namespace_packages(include=["reviews.*"]),
    version="0.1.0",
    license="MIT",
    description=("Code Review Manager written in Python. Standalone client included."),
    author="Kyle Harrison",
    author_email="kyle.harrison.dev@gmail.com",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/apoclyps/code-review-manager",
    download_url="https://github.com/apoclyps/code-review-manager/archive/0.1.0.tar.gz",
    keywords=["code review manager", "pull request review"],
    install_requires=_requires_from_file("requirements.txt"),
    entry_points={"console_scripts": ["reviews = cli:main"]},
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows :: Windows 10",
        "Programming Language :: Python :: 3.9",
    ],
)



from dataclasses import dataclass
from enum import Enum, unique


@unique
class Sound(Enum):
    """Enum used to provide a mapping to sound files."""

    SUCCESS = "audio/success.wav"
    FAILURE = "audio/failure.wav"


@unique
class Language(Enum):
    """Enum used to provide a mapping to icon files."""

    PYTHON = "icon/python_icon.png"
    RUBY = "icon/ruby_icon.png"
    GO = "icon/go_icon.png"
    JAVASCRIPT = "icon/javascript_icon.png"
    TERRAFORM = "icon/terraform_icon.png"


@dataclass
class PullRequestNotification:
    """Represents a pull request notification"""

    org: str
    repository: str
    name: str
    number: int

    def __repr__(self):
        return f"{self.org}/{self.repository} updated with {self.name}"



import sqlite3
from sqlite3 import Connection, Error
from typing import Any, List, Optional


class SQLClient:
    """create and manage database connections to the SQLite database"""

    database: str
    connection: Connection

    def __init__(
        self,
        path: str = "",
        filename: str = "",
        connection: Optional[Connection] = None,
    ) -> None:
        self.database = f"{path}/{filename}"
        if connection:
            self.connection = connection
        else:
            self.connection = self.create_connection()

    def create_connection(self) -> Connection:
        """create a database connection to the SQLite database specified by db_file."""
        return sqlite3.connect(self.database)

    def create_table(self, create_table_sql: str) -> None:
        """create a table from the create_table_sql statement."""

        cursor = self.connection.cursor()
        try:
            cursor.execute(create_table_sql)
        except Error as exc:
            print(exc)
            raise exc

    def query(self, sql: str, data=None) -> List[Any]:
        """query a table from a given sql statement."""

        with self.connection as db:
            cursor = db.cursor()

            cursor.execute("PRAGMA Foreign_Keys = ON")
            if data:
                cursor.execute(sql, data)
            else:
                cursor.execute(sql)

            return cursor.fetchall()

    def insert(self, sql: str, data=None) -> int:
        """query a table from a given sql statement."""

        with self.connection as db:
            cursor = db.cursor()

            cursor.execute("PRAGMA Foreign_Keys = ON")
            if data:
                cursor.execute(sql, data)
            else:
                cursor.execute(sql)

            return cursor.lastrowid
